import logging
from datetime import timedelta
from typing import Generator, List, Optional, Union  # for type hinting

# // from byteblowerll.byteblower import ConvertHTTPRequestStatusToString
# Helper functions to parse the strings (HTTP Method, ...)
# to the enumerations used by the API and vice versa
from byteblowerll.byteblower import HTTPClient  # for type hinting
from byteblowerll.byteblower import HTTPClientMobile  # for type hinting
from byteblowerll.byteblower import HTTPRequestStatus  # for type hinting
from byteblowerll.byteblower import HTTPServer  # for type hinting
from byteblowerll.byteblower import (
    HTTPServerStatus,
    ParseTCPCongestionAvoidanceAlgorithmFromString,
    TechnicalError,
)

from .._analysis.storage.tcp import TcpStatusData
from .._endpoint.endpoint import Endpoint
from .._endpoint.port import Port
from .._helpers.syncexec import SynchronizedExecutable  # for type hinting
from ..exceptions import (
    ByteBlowerTestFrameworkException,
    IncompatibleHttpServer,
    InvalidInput,
)
from ._http_client_controller import (
    EndpointTCPClientController,
    PortTCPClientController,
)
from .constants import HttpMethod, TCPCongestionAvoidanceAlgorithm
from .flow import RuntimeErrorInfo  # for type hinting
from .flow import Flow


class TcpFlow(Flow):
    """
    Flow, supporting multiple and/or restarting TCP clients.

    - Single HTTP server on the "WAN side" of the network.
    - One or multiple clients on the "CPE side" of the network.
    """

    __slots__ = (
        '_bb_tcp_server',
        '_bb_tcp_clients',
        '_http_method',
        '_applied_http_method',
        '_tcp_status_data',
        '_tcp_client_controller',
        '_client_port',
        '_server_port',
    )

    _CONFIG_ELEMENTS = Flow._CONFIG_ELEMENTS + ('http_method', )

    def __init__(
        self,
        source: Union[Port, Endpoint],
        destination: Union[Port, Endpoint],
        name: Optional[str] = None,
        http_method: HttpMethod = HttpMethod.AUTO,
        **kwargs,
    ) -> None:
        """
        Create a new TCP Flow.

        No clients, servers or sessions are created yet.
        """
        super().__init__(source, destination, name=name, **kwargs)
        self._bb_tcp_server: Optional[HTTPServer] = None
        self._bb_tcp_clients: Union[List[HTTPClient],
                                    List[HTTPClientMobile]] = []
        self._http_method = http_method
        self._applied_http_method = http_method
        self._tcp_status_data = TcpStatusData()

        self._tcp_client_controller: Union[PortTCPClientController,
                                           EndpointTCPClientController]

        self._server_port: Port
        self._client_port: Union[Port, Endpoint]

        # Sanity checks
        if self._http_method != HttpMethod.AUTO:
            raise NotImplementedError(
                f'TCP Flow {self.name!r} only supports'
                ' automatic HTTP method for now'
            )

    @property
    def http_method(self) -> str:
        """HTTP method used for HTTP (client) session."""
        return self._applied_http_method.value

    @property
    def finished(self) -> bool:
        """Returns True if the flow is done."""
        return self._tcp_client_controller.finished(
            self._client_port, self._bb_tcp_clients
        )

    @property
    def runtime_error_info(self) -> RuntimeErrorInfo:
        error_info = {}
        http_server_status = self._tcp_status_data._server_status
        if http_server_status == HTTPServerStatus.Error:
            # NOTE: No error message available for the server
            # server_error_message = (
            #     self._tcp_status_data._server_error_message
            # )
            server_error_message = "Failed to start HTTP Server"
            error_info['server_error_message'] = server_error_message
        client_error_messages = [
            client_error_message for (
                http_request_status,
                client_error_message,
            ) in self._tcp_status_data._client_status
            if http_request_status == HTTPRequestStatus.Error
        ]
        if client_error_messages:
            error_info['client_error_messages'] = client_error_messages
        return error_info

    def _set_tcp_server(
        self,
        server_port: Optional[Port] = None,
        tcp_port: Optional[int] = None,
        receive_window_scaling: Optional[int] = None,
        slow_start_threshold: Optional[int] = None,
        caa: Optional[TCPCongestionAvoidanceAlgorithm] = None,
    ) -> Optional[HTTPServer]:
        """Create a HTTP server.

        .. note::
           When a TCP port is given, an existing HTTP server can be re-used.

        :param server_port: Force HTTP server on the given ByteBlower Port.
           When set to ``None``, choose the port based on the :ref:HttpMethod,
           defaults to None
        :type server_port: Optional[Port], optional
        :param tcp_port: TCP port where the HTTP server listens to,
           defaults to None
        :type tcp_port: Optional[int], optional
        :param receive_window_scaling: When given, enable receive window
           scaling with the given scale factor, defaults to None
        :type receive_window_scaling: Optional[int], optional
        :param slow_start_threshold: TCP Slow start threshold value,
           defaults to None
        :type slow_start_threshold: Optional[int], optional
        :param caa: Use the given TCP congestion avoidance algorithm,
           defaults to None (server default)
        :type caa: Optional[TCPCongestionAvoidanceAlgorithm], optional
        :raises ByteBlowerTestFrameworkException: When a TCP server is
           already configured.
        :raises NotImplementedError: When no ``server_port`` is given and
           HttpMethod is not set to AUTO.
        :raises IncompatibleHttpServer: When an HTTP server is already
           configured with incompatible settings.
        :return: The new or existing HTTP Server. When it is already active,
           ``None`` is returned since we don't allow reconfiguring it
           once started.
        :rtype: Optional[HTTPServer]
        """
        # Sanity checks
        if self._bb_tcp_server:
            raise ByteBlowerTestFrameworkException('TCP server is already set')

        # Set the default client and server port
        if not server_port:
            if self._http_method != HttpMethod.AUTO:
                # TODO - Implement for given HTTP Method
                raise NotImplementedError(
                    f'TCP Flow {self.name!r} only support'
                    ' automatic HTTP method (for now)'
                )
            if self.source.is_natted:
                # Base Flow does not allow both source and destination NAT
                assert not self.destination.is_natted, (
                    'Source + destination natted is not supported'
                )
                server_port = self.destination
            else:
                server_port = self.source

        if tcp_port is not None:
            # Re-use existing HTTP Server.
            # Raise an error when incompatible settings are requested.
            for http_server in server_port.bb_port.ProtocolHttpServerGet():
                if http_server.PortGet() != tcp_port:
                    continue

                if receive_window_scaling is not None:
                    if not http_server.ReceiveWindowScalingIsEnabled():
                        raise IncompatibleHttpServer()
                    if (http_server.ReceiveWindowScalingValueGet()
                            != receive_window_scaling):
                        raise IncompatibleHttpServer()
                if (slow_start_threshold is not None
                        and (http_server.SlowStartThresholdGet()
                             != slow_start_threshold)):
                    raise IncompatibleHttpServer()
                if caa is not None and (
                        http_server.TcpCongestionAvoidanceAlgorithmGet()
                        != ParseTCPCongestionAvoidanceAlgorithmFromString(
                            caa.value)):
                    raise IncompatibleHttpServer()

                self._bb_tcp_server = http_server

                http_server_status = http_server.StatusGet()
                if http_server_status == HTTPServerStatus.Running:
                    # Do not return the HTTP server.
                    # The caller should not start or re-configure
                    # active HTTP servers.
                    return None
                return self._bb_tcp_server

        # Create a TCP server on the destination.
        http_server: HTTPServer = server_port.bb_port.ProtocolHttpServerAdd()
        if tcp_port is not None:
            http_server.PortSet(tcp_port)
        if receive_window_scaling is not None:
            http_server.ReceiveWindowScalingEnable(True)
            http_server.ReceiveWindowScalingValueSet(receive_window_scaling)
        if slow_start_threshold is not None:
            http_server.SlowStartThresholdSet(slow_start_threshold)
        if caa is not None:
            http_server.TcpCongestionAvoidanceAlgorithmSet(
                ParseTCPCongestionAvoidanceAlgorithmFromString(caa.value)
            )

        self._bb_tcp_server = http_server

        return self._bb_tcp_server

    def _set_tcp_client(
        self,
        client_port: Optional[Union[Port, Endpoint]] = None,
        server_port: Optional[Port] = None
    ):
        # Set the default client and server port
        if not client_port and not server_port:
            if self.source.is_natted:
                # Base Flow does not allow both source and destination NAT
                assert not self.destination.is_natted, (
                    'Source + destination natted is not supported'
                )
                logging.debug('%s: Server at destination', self.name)
                server_port = self.destination
                client_port = self.source
            else:
                logging.debug('%s: Server at source', self.name)
                server_port = self.source
                client_port = self.destination

            if isinstance(client_port, Port):
                self._tcp_client_controller = PortTCPClientController
            elif isinstance(client_port, Endpoint):
                self._tcp_client_controller = EndpointTCPClientController

        elif not server_port:
            raise InvalidInput(
                f'TCP Flow {self.name!r}: Client Port {client_port.name!r}'
                ' given without server Port'
            )
        elif not client_port:
            raise InvalidInput(
                f'TCP Flow {self.name!r}: Server Port {server_port.name!r}'
                ' given without client Port'
            )

        # NOTE: This check will also work when _client_port / _server_port
        #       were not configured yet:
        if (getattr(self, '_server_port', server_port) is not server_port or
                getattr(self, '_client_port', client_port) is not client_port):
            raise InvalidInput(
                'Using inconsistent HTTP Server/Client endpoints'
            )
        self._server_port = server_port
        self._client_port = client_port

        if self._http_method == HttpMethod.AUTO:
            if self._client_port == self.source:
                logging.debug('%s: Using PUT', self.name)
                self._applied_http_method = HttpMethod.PUT
            else:
                logging.debug('%s: Using GET', self.name)
                self._applied_http_method = HttpMethod.GET

    def _client_supports_tcp_parameters(self) -> bool:
        """Return whether the HTTP Client port supports setting TCP parameters.

        This is related to the TCP parameters which can
        be given to the :meth:`_add_client_session`.

        :return: Whether or not setting TCP parameters is supported.
        :rtype: bool
        """
        self._set_tcp_client()
        return self._tcp_client_controller.supports_tcp_parameters()

    def _add_client_session(
        self,
        client_port: Optional[Union[Port, Endpoint]] = None,
        server_port: Optional[Port] = None,
        request_duration: Optional[timedelta] = None,
        request_size: Optional[int] = None,
        maximum_bitrate: Optional[Union[int, float]] = None,
        ittw: Optional[timedelta] = None,
        receive_window_scaling: Optional[int] = None,
        slow_start_threshold: Optional[int] = None,
        caa: Optional[TCPCongestionAvoidanceAlgorithm] = None,
        tcp_port: Optional[int] = None,
        ip_traffic_class: Optional[int] = None,
    ) -> Generator[SynchronizedExecutable, None, Union[HTTPClient,
                                                       HTTPClientMobile]]:
        self._set_tcp_client()

        http_client_generator = self._tcp_client_controller.add_client_session(
            http_method=self._applied_http_method,
            client_port=client_port or self._client_port,
            server_port=server_port or self._server_port,
            bb_tcp_server=self._bb_tcp_server,
            request_duration=request_duration,
            request_size=request_size,
            maximum_bitrate=maximum_bitrate,
            ittw=ittw,
            receive_window_scaling=receive_window_scaling,
            slow_start_threshold=slow_start_threshold,
            caa=caa,
            tcp_port=tcp_port,
            ip_traffic_class=ip_traffic_class,
        )

        http_client = yield from http_client_generator

        # Add the returned HTTP Client instance
        self._bb_tcp_clients.append(http_client)

        return http_client

    def _last_client_session(self) -> Union[HTTPClient, HTTPClientMobile]:
        if not self._bb_tcp_clients:
            raise ByteBlowerTestFrameworkException('No TCP client created yet')
        return self._bb_tcp_clients[-1]

    def wait_until_finished(
        self, wait_for_finish: timedelta, result_timeout: timedelta
    ) -> None:
        """Wait until the flow finished traffic generation and processing.

        :param wait_for_finish: Time to wait for sessions closing
           and final packets being received.
        :type wait_for_finish: timedelta
        :param result_timeout: Time to wait for Endpoints to finalize
           and return their results to the Meeting Point.
        :type result_timeout: timedelta
        """
        self._tcp_client_controller.wait_until_finished(
            self._client_port, self._bb_tcp_clients, wait_for_finish,
            result_timeout
        )

    def stop(self) -> None:
        # 1. Stop all TCP clients
        self._tcp_client_controller.stop(self._bb_tcp_clients)

        # 2. Stop TCP Server
        if self._bb_tcp_server is not None:
            self._bb_tcp_server.Stop()

        super().stop()

    def analyse(self) -> None:
        self._tcp_status_data._server_status = self._bb_tcp_server.StatusGet()
        # NOTE: No error message available for the server
        # self._tcp_status_data._server_error_message = (
        #     self._bb_tcp_server.ErrorMessageGet()
        # )
        for tcp_client in self._bb_tcp_clients:
            try:
                http_request_status: HTTPRequestStatus = (
                    tcp_client.RequestStatusGet()
                )
            except TechnicalError as e:
                # NOTE: HTTPClientMobile.RequestStatusGet() is not implemented
                #       (as for API v2.21)
                http_request_status = None
                logging.warning(
                    "Unable to get HTTPRequest status: %s", e.getMessage()
                )
            client_error_message: str = tcp_client.ErrorMessageGet()
            self._tcp_status_data._client_status.append(
                (http_request_status, client_error_message)
            )

        super().analyse()

    def _release(self) -> None:
        client_port = self._client_port
        server_port = self._server_port

        try:
            bb_tcp_clients = self._bb_tcp_clients
            del self._bb_tcp_clients
        except AttributeError:
            logging.warning(
                'TcpFlow: TCP clients already destroyed?', exc_info=True
            )
        else:
            self._tcp_client_controller.release(client_port, bb_tcp_clients)
        try:
            del self._client_port
        except AttributeError:
            logging.warning(
                'TcpFlow: TCP client already removed?', exc_info=True
            )

        # ! FIXME: Don't destroy when re-used existing HTTP Server !
        try:
            bb_tcp_server = self._bb_tcp_server
            del self._bb_tcp_server
        except AttributeError:
            logging.warning(
                'TcpFlow: TCP server already destroyed?', exc_info=True
            )
        else:
            if bb_tcp_server is not None:
                server_port.bb_port.ProtocolHttpServerRemove(bb_tcp_server)
        try:
            del self._server_port
        except AttributeError:
            logging.warning(
                'TcpFlow: TCP server already removed?', exc_info=True
            )
